Require Import Arith.

Inductive even : nat -> Prop :=
  | O_even : even 0
  | plus_2_even : forall n:nat, even n -> even (S (S n)).


Hint Resolve O_even plus_2_even.

Fixpoint mult2 (n:nat) : nat :=
  match n with
  | O => 0
  | S p => S (S (mult2 p))
  end.

Lemma mult2_even : forall n:nat, even (mult2 n).
Proof. 
induction n; simpl; auto.
Qed.

Theorem sum_even : forall n p:nat, even n -> even p -> even (n + p).
Proof.
intros n. elim n. simpl. intros. assumption. intros. simpl.
Restart.
 intros n p Heven_n. elim Heven_n. simpl. trivial.
intros.
simpl.
apply plus_2_even.
apply H0. assumption.
Restart.
intros n p Heven_n; elim Heven_n; simpl; auto.
Qed.

Hint Resolve sum_even.

Lemma square_even : forall n:nat, even n -> even (n * n).
Proof.
 intros n Hn; elim Hn; simpl; auto.
 Check mult_comm.
 intros n0 H0 H1. rewrite (mult_comm n0 (S (S n0))).
 simpl.
apply plus_2_even.
 apply sum_even. assumption. apply plus_2_even.  apply sum_even. assumption. 
 apply sum_even. assumption.
 apply sum_even. assumption. assumption.
Restart.
intros n Hn; elim Hn; simpl; auto.
intros n0 H0 H1; rewrite (mult_comm n0 (S (S n0))). 
simpl. 
right.
apply sum_even; auto.
Qed.


Lemma even_mult2 : forall n:nat, even n -> (exists p, n = mult2 p).
Proof.
intros.
induction  H.
exists 0.
simpl. reflexivity.
case IHeven. intros. exists (S x); simpl; auto.
Restart.
 induction 1.
 exists 0;simpl;auto.
 case IHeven. intros p Hp.
 exists (S p); simpl; auto.
Qed.

Lemma even_sum: forall n m:nat, even n -> even m -> even(n+m).
Proof.
intros n p H; elim H. 
trivial.
intros n0 E0 Hrec Heven; simpl.
apply plus_2_even; auto.
Qed.


Theorem not_even_1 : ~even 1.
Proof.
unfold not. intros H.  (* elim H.*)
Check even_ind.
 inversion H.
Qed.

Theorem plus_2_even_inv : forall n:nat, even (S (S n))-> even n.
Proof.

 intros n H; inversion H.
 assumption.
Qed.

Theorem not_even_1' : ~even 1.
Proof.
 intro H.
 generalize (refl_equal 1).
 pattern 1 at -2.
 elim H.
 discriminate. 
 discriminate 3.
Qed.

Theorem plus_2_even_inv' : forall n:nat, even (S (S n))-> even n.
Proof.
 intros n H.
 generalize (refl_equal (S (S n))); pattern (S (S n)) at -2.
 elim H.
 intros H1; discriminate H1.
 intros n0 H'0 H' H'1.
 injection H'1; intros H'2; rewrite <- H'2; assumption.
Qed.

