Require Import List.
Require Export ZArith.
Require  Export ZArith_dec.
Require Extraction.



Local Open Scope Z_scope.

Print Z_dec.

Set Implicit Arguments.

Inductive sorted : list Z -> Prop :=
| sorted0: sorted nil
| sorted1 : forall z, sorted (z::nil)
| sorted2 : forall z1 z2 l, z1 <= z2 -> sorted (z2::l) -> sorted (z1::z2::l).

Hint Resolve sorted0 sorted1 sorted2 : sort.

(* Principio de inversao para uma lista ordenada *)
Theorem sorted_inv : forall z l, sorted (z::l) -> sorted l.
Proof.
  intros z l H.
  inversion H.
  constructor.
  assumption.
Qed.

Print sorted_inv.

Print Z.eq_dec.

Check Z_le_gt_dec.

(* Número de ocorrencias de um inteiro numa lista *)

Fixpoint nb_occ (z:Z) (l:list Z) {struct l} : nat :=
  match l with
    | nil => O
    | (z'::l') => match Z.eq_dec z  z' with
                    | left  _ => S (nb_occ z l')
                    | right _ => nb_occ z l'
                  end
  end.

Extraction Language Haskell.

Extraction nb_occ. 

(* Duas listas sao equivalentes se tiverem exactamente os mesmos elementos *)

Definition equiv (l l':list Z) := forall z, nb_occ z l = nb_occ z l'.


(* A equivalencia e reflexiva *)
Lemma equiv_refl : forall l, equiv l l.
Proof.
  intro.
  unfold equiv.
  trivial.
Qed.

(* A equivalencia e transitiva *)
Lemma equiv_trans : forall l l' l'', equiv l l' -> equiv l' l'' ->  equiv l l''.
Proof.
  unfold equiv.
  intros l l' l'' H1 H2.
  intro.
  rewrite (H1 z).
  apply H2.
Qed.

(* Equivalencia mantem-se se adicionado o mesmo elmento a ambas as listas *)
Lemma equiv_cons : forall z l l', equiv l l' -> equiv (z::l) (z::l').
Proof.
  intros.
  unfold equiv.
  simpl.
  intro.
  Check Z.eq_dec.
  case(Z.eq_dec z0 z).
  intro.
  unfold equiv in H.
  rewrite H.
  reflexivity.
  intro.
  unfold equiv in H.
  apply H.
Qed.

(* Equivalencia mantem-se no caso de permutacoes de uma mesma lista *)
Lemma equiv_perm : forall a b l l', equiv l l' -> equiv (a::b::l) (b::a::l').
Proof.
intros.
unfold equiv;simpl.
intro.
case(Z_eq_dec z a).
intro.
case(Z_eq_dec z b).
intro.
unfold equiv in H.
rewrite (H z).
reflexivity.
intro.
rewrite (H z).
reflexivity.
intro.
rewrite (H z).
reflexivity.
Qed.

Hint Resolve equiv_cons equiv_refl equiv_perm : sort.

Check Z_le_gt_dec.

(* Funcao que insere um inteiro na sua correcta posicao, numa determinada lista *)
Fixpoint ins_aux (z:Z) (l:list Z) {struct l} : list Z :=
match l with
| nil => z::nil
| (a::l') => match Z_le_gt_dec z a with
                | left  _ => z::a::l'
                | right _ => a::(ins_aux z l')
              end
end.

Extraction ins_aux.

(* A insercao nao altera a equivalencia de uma lista *)
Lemma ins_aux_equiv : forall l x, equiv (x::l) (ins_aux x l).
Proof.
induction l as [|a l0 H]; simpl; auto with sort.
intros x;case(Z_le_gt_dec x a).
simpl;auto with sort.
intro;apply equiv_trans with (a::x::l0);auto with sort.
Qed.


(* A insercao manter a ordem de uma lista *)
Lemma ins_aux_sorted : forall l x, sorted l -> sorted (ins_aux x l).
Proof.
  intros.
  elim H.
  simpl.
  constructor.
  intros.
  simpl.
  case(Z_le_gt_dec x z);simpl;auto with sort zarith.
  intros z1 z2.
  simpl.
  case(Z_le_gt_dec x z2);simpl;intros;case(Z_le_gt_dec x z1);simpl;auto with sort zarith.
Qed.
  
Print ins_aux_sorted.

Lemma my_sort : forall l, {l' : list Z | equiv l l' /\ sorted l'}.
Proof.
  intros.
  induction l as [| a l H1].
  exists nil.
  split;auto with sort.
  case H1;intros l' [H2 H3].
  exists (ins_aux a l').
  split.
  apply equiv_trans with (a::l');auto with sort.
  apply (ins_aux_equiv).
  apply (ins_aux_sorted).
  assumption.
Qed.

Extraction my_sort.
Extraction ins_aux.

Recursive Extraction my_sort.
